import argparse
from fish_tool import register
import os
from fish_tool.log_tool import logs

parser = argparse.ArgumentParser(description='AI实验室')

parser.add_argument('config_name', type=str, help='配置名称')
parser.add_argument('--folder', type=str, default='', help='模型存储位置')

parser.add_argument('--start', type=int, default=0, help='训练起始位置（闭区间）')
parser.add_argument('--end', type=int, default=900000000, help='训练结束位置（开区间）')
parser.add_argument('--repeat', type=int, default=1, help='训练数据重复次数')


def root(*path):
    out = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../', *path))
    return out


def train(config_name, folder='', start=0, end=900000000, repeat=1):
    # 训练函数（执行训练流程）
    # 注意： 必须先import注册对象所在的模块，才能正常使用register （使用register.walk_import)
    config = register.get(config_name)()
    if folder:
        config.folder = root(folder)
    else:
        config.folder = root(f'data/{config_name}')
    logs.reset(config.folder)
    logs.brief.info(f'name={config_name}   config={config}')
    Learner = register.get(config.learner_name)
    learner = Learner(config)
    learner.train(start=start, end=end, repeat=repeat)


def cmd_train():
    # 通过命令行参数 执行训练函数
    args = parser.parse_args()
    config_name = args.config_name
    folder = args.folder
    train(config_name, folder, start=args.start, end=args.end, repeat=args.repeat)



import sys
from glob import glob
import datetime
from copy import deepcopy
from types import MethodType, FunctionType
import inspect
import os
import json
import hashlib
import uuid
import base64
import pickle
import signal
import time
import random
from fish_tool.log_tool import logs, InfoError
from fish_tool.db.sqlite_tool import SqliteTool


def sys_path_add(*paths, logs=None):
    d = os.path.abspath(os.path.join(*paths))
    if not os.path.isdir(d):
        d = os.path.dirname(d)
    if d not in sys.path:
        sys.path.insert(0, d)
        if logs is not None:
            logs.tmp.info(f'sys.path.insert("{d}")')


def get_cache_scroll_path(folder, name, max_num=5, add_one=True, show_log=False):
    # 获取滚动缓存文件名 如果文件数量达到max-num则自动删除最老的文件
    os.makedirs(folder, exist_ok=True)
    ptn = os.path.join(folder, f'*_{name}')
    paths = glob(ptn)
    pres = []
    for path in paths:
        pre = int(os.path.basename(path).split('_')[0])
        pres.append(pre)
    pres.sort()
    if len(pres) >= max_num:
        old_path = os.path.join(folder, f'{pres[0]}_{name}')
        os.remove(old_path)
        if show_log:
            logs.brief.info(f'delete {old_path}')
    if pres:
        pre = pres[-1]
        if add_one:
            pre += 1
    else:
        pre = 0
    path = os.path.join(folder, f'{pre}_{name}')
    return path


class DelayKeyboardInterrupt(object):
    # Ctrl+C 延迟关闭程序的上下文管理器， 例如避免写入文件时被关闭导致文件损坏
    # 使用方法 with DelayKeyboardInterrupt():
    #             save_code()
    def __enter__(self):
        self.signal_received = False
        # signal.SIGINT  键盘中断（如break键被按下）  （暂时没有支持kill 进程发送的15信号）
        self.old_handler = signal.signal(signal.SIGINT, self.handler)

    def handler(self, sig, frame):
        self.signal_received = (sig, frame)
        logs.brief.info('SIGINT received. Delaying KeyboardInterrupt.')

    def __exit__(self, type, value, traceback):
        signal.signal(signal.SIGINT, self.old_handler)
        if self.signal_received:
            self.old_handler(*self.signal_received)


def obj_to_sort_str(obj):
    # 把字典或数组都sort后再生成str （避免字典或者集合里的元素顺序不一致，导致str不一致）
    if isinstance(obj, set):
        txt = ', '.join(sorted(obj_to_sort_str(t) for t in obj))
        return '{' + txt + '}'
    elif isinstance(obj, dict):
        txt = ', '.join(f'{k}:{obj_to_sort_str(v)}' for k, v in sorted(obj.items()))
        return '{' + txt + '}'
    elif isinstance(obj, list):
        txt = ', '.join(obj_to_sort_str(t) for t in obj)
        return '[' + txt + ']'
    elif isinstance(obj, tuple):
        txt = ', '.join(obj_to_sort_str(t) for t in obj)
        return '(' + txt + ')'
    elif isinstance(obj, str):
        return f'"{obj}"'
    else:
        return str(obj)


_type__str = {  # 类型 对应的 类型描述文本
    type(1): 'int',
    type(0.1): 'float',
    type('a'): 'str',
    type(True): 'bool',
}


def diff_struct(o1, o2):
    if type(o1) != type(o2):
        return f'type 不同 {type(o1)} <> {type(o2)}'

    if isinstance(o1, dict):
        out = {}
        k1s = set(o1)
        k2s = set(o2)
        for k in k1s - k2s:
            out[k] = 'object1 多余的key'
        for k in k2s - k1s:
            out[k] = 'object2 多余的key'
        for k in k1s & k2s:
            diff = diff_struct(o1[k], o2[k])
            if diff:
                out[k] = diff
        return out
    if isinstance(o1, (list, tuple)):
        min_num = min(len(o1), len(o2))
        max_num = max(len(o1), len(o2))
        out = []
        for i in range(min_num):
            diff = diff_struct(o1[i], o2[i])
            if diff:
                out.append(diff)
        if len(o1) > len(o2):
            long = o1
            long_name = 'object1'
        else:
            long = o2
            long_name = 'object2'
        for i in range(min_num, max_num):
            out.append({f'{long_name}[{i}] {type(o1)}多余的元素': long[i]})
        return out
    if isinstance(o1, set):
        out = {}
        for item in o1 - o2:
            out.setdefault(f'object1 set()多余的元素', []).append(item)
        for item in o2 - o1:
            out.setdefault(f'object2 set()多余的元素', []).append(item)
        return out
    # if isinstance(o1, (int, float, bool, str)):  获取包括其他类型
    if o1 != o2:
        return f'value 不同 {o1} <> {o2}'


def obj_to_struct(obj, str_max_len=10, list_max_num=3, return_value=False):
    # 返回对象的结构  str_max_len是文本显示最大长度   list_max_num数组、元组、集合最多查看的数量上上限
    # return_value 为真时int、float、str、bool返回具体的值   为假时返回类型描述文本
    if isinstance(obj, dict):
        return {k: obj_to_struct(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple, set)):
        tmp = set()
        out = []
        for sub_obj in list(obj)[:list_max_num]:
            sub_struct = obj_to_struct(sub_obj)
            sub_txt = obj_to_sort_str(sub_struct)
            if sub_txt not in tmp:
                out.append(sub_struct)
                tmp.add(sub_txt)
        return type(obj)(out)
    elif obj is None:
        return None
    elif isinstance(obj, (int, float, bool)):
        if return_value:
            return obj
        else:
            return _type__str[type(obj)]
    elif isinstance(obj, str):
        if return_value:
            return obj[:str_max_len]
        else:
            return _type__str[type(obj)]
    else:
        return str(type(obj))


def lazy_property(func):
    # 给类的属性添加懒加载模式   使用方法，在类的属性方法上添加@lazy_property
    attr_name = "_lazy_" + func.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            logs.brief.info(f'计算懒加载属性 {func}')
            setattr(self, attr_name, func(self))
        return getattr(self, attr_name)

    return _lazy_property


def cache_hour_property(hour=1):
    ms = hour * 60 * 60 * 1000

    def decorator(func):
        # 给类的属性添加缓存模式(间隔时间＜hour时直接使用缓存数据)   使用方法，在类的属性方法上添加 @cache_hour_property(间隔小时)
        value_name = "_cache_" + func.__name__
        time_name = '_time_' + func.__name__

        @property
        def _cache_property(self):
            now = get_timestamp()
            if not hasattr(self, time_name) or (now - getattr(self, time_name)) > ms:
                logs.rotate.info(f'计算小时缓存属性 {func}')
                setattr(self, value_name, func(self))
                setattr(self, time_name, now)
            return getattr(self, value_name)

        return _cache_property

    return decorator


def singleton(cls):
    # 单例模式的类装饰器    在类的上一行加上 @singleton （不支持带初始化参数的类）
    _instance = {}

    def inner():
        if cls not in _instance:
            _instance[cls] = cls()
        return _instance[cls]

    return inner


class ProcessMemory:
    pre_mem = 'mem_'  # 对应数据表里记录该条记录是否已经记录
    pre_result = 'res_'  # 对应数据表里记录数据结果

    def __init__(self, path):
        self.mem_path = path
        self.last_save_time = int(time.time() * 1000)
        self.mem = SqliteTool(path)
        self.fn_set = set()
        tables = self.mem.get_table_names()
        for table in tables:
            if table.startswith(self.pre_mem):
                num = self.mem.num(table)
                logs.tmp.info(f'{table} 数量={num}')
                self.fn_set.add(table[4:])

    def _info_to_key(self, info):
        if isinstance(info, (str, int)):
            return str(info)
        elif isinstance(info, dict):
            infos = [f'{k}={v}' for k, v in info.items()]
            infos.sort()
            return ','.join(infos)
        else:
            raise InfoError(f'未支持的类型 ({type(info)})info={info}')

    def init_fn(self, fn):
        if fn not in self.fn_set:
            table = f'{self.pre_mem}{fn}'
            self.mem.create_table(table, {'key': 'str', 'time': 'int'}, uni_keys=['key'], primary_keys=['key'])
            table = f'{self.pre_result}{fn}'
            self.mem.create_table(table, {'key': 'str', 'value': 'str', 'time': 'int'}, uni_keys=['key'], primary_keys=['key'])
            self.fn_set.add(fn)

    # 记录某个函数对某个数据的操作时间（秒级时间戳）
    def add_record(self, fn, info, result=None):
        # fn是函数名str         info可以是str，也可以是dict
        self.init_fn(fn)
        process_time = get_timestamp()
        key = self._info_to_key(info)
        self.mem.insert(f'{self.pre_mem}{fn}', {'key': key, 'time': process_time})
        if result is not None:
            self.mem.insert(f'{self.pre_result}{fn}', {'key': key, 'time': process_time, 'value': result})

    # 判断某个函数（在start_time以后的时间内）是否处理过某条数据
    def is_processed(self, fn, info, start_time=0):
        if fn not in self.fn_set:
            return False
        key = self._info_to_key(info)
        condition = f'key="{key}"'
        lines = self.mem.select(f'{self.pre_mem}{fn}', condition)
        if lines:
            process_time = lines[0]['time']
            return process_time > start_time
        else:
            return False

    def get_result(self, fn, info):
        key = self._info_to_key(info)
        condition = f'key="{key}"'
        lines = self.mem.select(f'{self.pre_mem}{fn}', condition)
        if lines:
            return lines[0]['value']
        else:
            return None

    # 主动触发把记忆存到pkl文件中
    def save(self):
        self.mem.commit()
        logs.brief.info(f'memory saved in {self.mem_path}')


def trans_excel(excel_path):
    # 每个表格的第一行是表头，下面都是数据， 转化成字典 {'表名': [{'键名': 值}]}
    import openpyxl
    from tqdm import tqdm
    table__lines = {}
    wb = openpyxl.load_workbook(excel_path, read_only=True, data_only=True)
    for sheet in wb:
        table = sheet.title
        data = sheet['A1': 'XXX500000']
        lines = []
        idx__field = {}
        for i, cell in enumerate(data[0]):
            field = cell.value
            if field:
                field = field.repalce(' ', '')
                if field:
                    field = field.lower()
                    idx__field[i] = field
            for line in tqdm(data[1:]):
                one = {}
                for i, field in idx__field.items():
                    v = line[i].value
                    if v:
                        one[field] = v
                if one:
                    lines.append(one)
            logs.brief.info(f'{table} 数据量={len(lines)}')
        table__lines[table] = lines
    return table__lines


def infos_find(infos, key, value):
    # infos = [{'key': value}]  在字典数组里找到kv一致的字典返回
    for i, info in enumerate(infos):
        if info.get(key) == value:
            return i, info
    return None, None


def differ_days(time_a, time_b):
    """
    计算日期相差天数  time_a/time_b 都是格式=“xxxx-xx-xx”的文本
    注意 如果time_a < time_b 则会返回负数
    """
    a = time_a.split('-')
    b = time_b.split('-')
    d1 = datetime.date(int(a[0]), int(a[1]), int(a[2]))
    d2 = datetime.date(int(b[0]), int(b[1]), int(b[2]))
    return (d1 - d2).days


def get_timestamp(txt='', fmt="%Y-%m-%dT%H:%M:%S"):
    # 返回毫秒时间戳（浮点数)   # txt = '2016-05-05 20:28:54'   T是标准日期+时间的表示格式
    if not txt:
        return int(time.time() * 1000)
    time_array = time.strptime(txt, fmt)
    return int(time.mktime(time_array) * 1000)


def get_timetxt(timestamp=None, fmt='%Y-%m-%dT%H:%M:%S', delta_day=0, delta_ms=0):
    # 返回时间文本   delta增量时间   # timestamp = 1462451334000   T是标准日期+时间的表示格式
    if timestamp is None:
        timestamp = time.time() * 1000
    timestamp += delta_day * 24 * 60 * 60 * 1000
    timestamp += delta_ms
    time_local = time.localtime(timestamp / 1000)
    return time.strftime(fmt, time_local)


def get_begin_timestamp(delta_day=0, day=''):
    # 获取某一天的0点的时间戳    delta是增量天数   如果day为空文本则获取当天 文本格式="2021-01-01"
    if not day:
        day = get_timetxt(fmt='%Y-%m-%d')
    out = get_timestamp(f'{day}T00:00:00')
    out += delta_day * 24 * 60 * 60 * 1000
    return out


def return_dict(obj, *keys):
    out = {}
    for key in keys:
        out[key] = obj.get(key)
    return out


def uid4():
    # 基于随机数，有一定的重复概率
    return str(uuid.uuid4()).replace('-', '')


def sha256(txt):
    return hashlib.sha256(txt.encode('utf8')).hexdigest()


def md5(txt):
    out = hashlib.md5(txt.encode('utf8')).hexdigest()
    return str(out)


def hash_url(url):
    # 把 文件的url转成哈希值的文件名
    tail = url.split('/')[-1]
    exts = tail.split('.')
    if len(exts) > 1:
        ext = f'.{exts[-1]}'
    else:
        ext = ''
    md5 = hashlib.md5(url.encode(encoding='UTF-8')).hexdigest()
    name = f'{md5}{ext}'
    return name


def hash_file(binfile):
    """该函数返回传入文件的SHA-1哈希值   输入文件二进制流：open(path, 'rb').read()"""
    # 创建一个哈希对象
    h = hashlib.sha1()
    h.update(binfile)
    # 返回摘要的十六进制表示
    return h.hexdigest()


def txt_to_base64(txt):
    # 编码
    encode_str = base64.encodebytes(txt.encode('utf8'))  # b'aGVsbG8gd29ybGQh\n'
    txt = encode_str.decode('utf8')
    txt = txt.strip()
    return txt


def base64_to_txt(msg):
    try:
        decode_str = base64.decodebytes(msg.encode('utf8'))  # b'hello world!'
        return decode_str.decode('utf8')
    except:
        return None


def make_token(user_id, seed, sha_len):
    user_id = str(user_id)
    sha = sha256(user_id + seed)
    user_64 = txt_to_base64(user_id)
    token = sha[:sha_len] + user_64
    return token


def check_token(token, seed, sha_len):
    # logs.tmp.info(f'check_token(token={token}, seed={seed}, sha_len={sha_len})')
    head = token[:sha_len]
    # logs.tmp.info(f'    head={head}')
    user_id = base64_to_txt(token[sha_len:])
    # logs.tmp.info(f'    user_id={user_id}')
    if user_id:
        sha = sha256(user_id + seed)
        # logs.tmp.info(f'    whole sha={sha}     pre={sha[:sha_len]}')
        if head == sha[:sha_len]:
            return {'ok': True, 'user_id': user_id}
        else:
            return {'ok': False, 'detail': '不一致', 'user_id': ''}
    else:
        return {'ok': False, 'detail': '解码失败', 'user_id': ''}


def cut_data(data, test_num=50, old_split_path=None):
    # 切分数据为训练集、验证集、测试集   （如果有老的切分信息，则使用老的验证集和测试集的_id）
    if old_split_path and os.path.isfile(old_split_path):
        split_info = json.load(open(old_split_path))
        valid_ids = set(split_info['valid'])
        test_ids = set(split_info['test'])
        for dtype in ['train', 'valid', 'test']:
            logs.brief.info(f'split_info: {dtype} data has {len(split_info[dtype])}')
        train, valid, test = [], [], []
        for doc in data:
            if doc['_id'] in valid_ids:
                valid.append(doc)
            elif doc['_id'] in test_ids:
                test.append(doc)
            else:
                train.append(doc)
    else:
        random.shuffle(data)
        logs.brief.info(f'未提供split_info')
        valid = data[:test_num]
        test = data[test_num:test_num + test_num]
        train = data[test_num + test_num:]
    out = {'train': train, 'valid': valid, 'test': test}
    split_info = {'num': {}}
    for dtype in ['train', 'valid', 'test']:
        split_info[dtype] = []
        split_info['num'][dtype] = len(out[dtype])
        logs.brief.info(f'{dtype} data has {len(out[dtype])}')
        for doc in out[dtype]:
            split_info[dtype].append(doc['_id'])
    out['split_info'] = split_info
    return out


def code_path(obj):
    try:
        filename = inspect.getfile(obj)
        line_no = inspect.getsourcelines(obj)[-1]
    except:
        return obj.__name__
    return f'File "{filename}", line {line_no} >> {obj.__name__}'


def get_safe_args(fn, raw_args):
    # 获取一个函数能够传入的参数（去掉多余的值） 避免多余的参数导致报错
    if isinstance(fn, (FunctionType, MethodType)):
        fn = fn
    else:  # 如果是类，则fn=对象的初始化函数
        fn = fn.__init__
    full_args = inspect.getfullargspec(fn)
    if full_args.varargs:
        raise ValueError(f'{fn} 不支持数组参数={full_args.varargs}')
    if full_args.varkw:  # 具有字典参数时，可以输入任意参数，所以全部返回
        return raw_args
    else:
        return {k: v for k, v in raw_args.items() if k in full_args.args}


def save_score_to_html(score, path, width='1000px', height='280px'):
    import pyecharts.globals
    from pyecharts import options as opts
    page = pyecharts.charts.Page()
    labels = list(score['best']['train'].keys())
    keys = [str(i) for i in range(len(score['history']))]
    line_opt = opts.InitOpts(theme=pyecharts.globals.ThemeType.CHALK, width=width, height=height)
    if 'loss' in score['best']:
        line = pyecharts.charts.Line(init_opts=line_opt)
        line.set_global_opts(title_opts=opts.TitleOpts(title='train-loss'), legend_opts=opts.LegendOpts())
        line.add_xaxis(keys)
        for label in labels:
            values = [round(t['loss'], 6) for t in score['history']]
            line.add_yaxis(label, values, is_symbol_show=False)
        page.add(line)

    for dtype in ['train', 'valid', 'test']:
        line = pyecharts.charts.Line(init_opts=line_opt)
        line.set_global_opts(title_opts=opts.TitleOpts(title=f'性能指标-{dtype}'), legend_opts=opts.LegendOpts())
        line.add_xaxis(keys)
        for label in labels:
            values = [round(t[dtype][label], 4) for t in score['history']]
            line.add_yaxis(label, values, is_symbol_show=True)
        page.add(line)
    page.render(path)


def sleep_waiting(total=10, desc=''):
    # 失眠等待N秒 同时打印出提示文本
    for i in range(1, total + 1):
        print(f'\rsleep_waiting: {i}/{total}  {desc}', end='')
        time.sleep(1)


def weight_choice(weights, items=None):
    # 按权重随机返回数组的一个元素  （items是数组内容， 如果不提供则返回元素的index）
    items = items or list(range(len(weights)))
    return random.choices(items, weights=weights, k=1)[0]


def weight_sample(k, weights, items=None):
    # (不重复的）按权重采样k个元素  （items是数组内容， 如果不提供则返回元素的index）
    weights = deepcopy(weights)  # deepcopy避免改变输入的参数内容
    items = deepcopy(items) or list(range(len(weights)))
    if len(items) <= k:
        return items
    out = []
    for _ in range(k):
        index = weight_choice(weights)
        out.append(items.pop(index))
        weights.pop(index)
    return out


def recursion_round_dict(data, ndigits=2):
    # 递归的把字典里的浮点数四舍五入（便于查看时候方便）
    out = {}
    for k, v in data.items():
        if isinstance(v, float):
            out[k] = round(v, ndigits)
        elif isinstance(v, dict):
            out[k] = recursion_round_dict(v, ndigits)
        else:
            out[k] = v
    return out


def cache_glob(ptn, cover=False):
    # 有一些 glob递归读取太慢了， 放在缓存里面   如果cover=True则读取新的glob结果覆盖缓存
    if os.path.exists('C:/'):
        d = 'C:/.fish_tool_cache'
    else:
        d = '~/.fish_tool_cache'
    d = os.path.abspath(d)
    os.makedirs(d, exist_ok=True)
    name = sha256(ptn) + '.pkl'
    path = os.path.join(d, name)
    if cover or not os.path.exists(path):
        logs.tmp.info(f'get new glob({ptn}) path={path}')
        out = sorted(glob(ptn))  # 排序后可以固定顺序
        with open(path, 'wb') as f:
            pickle.dump(out, f)
    else:
        logs.tmp.info(f'get cache glob({ptn}) path={path}')
        out = pickle.load(open(path, 'rb'))
    return out


def read_jsonlist(path):
    data = []
    for line in open(path, encoding='utf8').readlines():
        line = line.strip('\n').strip(' ')
        if line:
            data.append(json.loads(line))
    return data


def write_jsonlist(data, path):
    f = open(path, 'w', encoding='utf8')
    for doc in data:
        txt = json.dumps(doc, ensure_ascii=False)
        f.write(f'{txt}\n')
    f.close()



import inspect
import importlib
import pkgutil
from fish_tool.log_tool import logs, InfoError

__registry = {}

import_cache = set()


def __walk_import(package_name):
    if package_name not in import_cache:
        import_cache.add(package_name)
        module = importlib.import_module(package_name)
        path = getattr(module, '__path__', [])
        if path:
            for finder, name, is_package in pkgutil.walk_packages(path, prefix=package_name + '.'):
                __walk_import(name)


def walk_import(package_name):
    # 递归载入一个包的全部子包和子模块
    logs.tmp.info(f'walk_import {package_name}')
    __walk_import(package_name)


# def add(name):
#     # 装饰器：新增注册
#     def decorator(cls):
#         if name in __registry:
#             if cls.__module__ == '__main__' or __registry[name].__module__ == '__main__':
#                 # 在某个模块下直接执行 会重复注册, 避免第二次时报错
#                 return cls
#             if len(cls.__module__) < len(__registry[name].__module__):
#                 # 如果短的模块名字属于长的模块名字， 说明是重复的
#                 short = cls.__module__
#                 long = __registry[name].__module__
#             else:
#                 short = __registry[name].__module__
#                 long = cls.__module__
#             if short in long:
#                 return cls
#             if cls == __registry[name]:
#                 return cls
#             message = f'"{cls}"的名称"{name}"已经被"{__registry[name]}"注册'
#             raise InfoError(message)
#         else:
#             __registry[name] = cls
#         return cls
#
#     return decorator


def add(name):
    # 装饰器：新增注册
    def decorator(cls):
        if name in __registry:
            if __registry[name] == cls:
                return cls
            old = code_path(__registry[name])
            new = code_path(cls)
            message = f'名称="{name}  重复注册\n{old}\n{new}'
            raise InfoError(message)
        else:
            __registry[name] = cls
        return cls

    return decorator


def get(name: str):
    # 通过名字获取类
    if name not in __registry:
        raise InfoError(f'"{name}" 未注册')
    return __registry.get(name)


def code_path(obj):
    try:
        filename = inspect.getfile(obj)
        lineno = inspect.getsourcelines(obj)[-1]
    except:
        return obj.__name__
    return f'File "{filename}", line {lineno} >> {obj.__name__}'


def show():
    # 显示已经注册的对象
    items = [f'\t{code_path(subclass): <100} 名称: {k}' for k, subclass in __registry.items()]
    items_txt = '\n'.join(items)
    head = '=' * 43 + ' 注册的类 ' + '=' * 43
    tail = '=' * 130
    logs.tmp.info(f'{head}\n{items_txt}\n{tail}')



import torch
import sklearn.metrics


def get_f1(tp, fp, fn):
    if tp + fn + fp == 0:
        return -1
    if tp + fp == 0:
        precision = 0
    else:
        precision = tp / (tp + fp)
    if tp + fn == 0:
        recall = 0
    else:
        recall = tp / (tp + fn)
    if precision + recall < 0.000001:
        f1 = 0
    else:
        f1 = 2. * ((precision * recall) / (precision + recall))
    return f1


class MultiHotF1Scorer:
    # 计算多标签的F1
    def __init__(self):
        self.pred = []
        self.gold = []
        self.average = 'macro'

    def update(self, pred, gold):
        if isinstance(pred, torch.Tensor):
            pred = pred.long().tolist()
        if isinstance(gold, torch.Tensor):
            gold = gold.long().tolist()

        if pred:
            if not isinstance(pred[0], list):
                pred = [pred]  # 确保pred~~（bsz， doc_num）
                gold = [gold]
            for pred_line, gold_line in zip(pred, gold):
                self.pred.extend(pred_line)
                self.gold.extend(gold_line)

    def f1(self, reset=True):
        value = sklearn.metrics.f1_score(self.gold, self.pred, average=self.average)
        if reset:
            self.pred = []
            self.gold = []
        return value


class MultiLabelF1Scorer:
    # 多标签任务的F1   计算每个类别和总F1
    def __init__(self, labels):
        assert len(labels) > 0
        self.label_num = len(labels)
        self.label__scorer = {label: MultiHotF1Scorer() for label in labels}

    def update(self, label, pred, gold):
        self.label__scorer[label].update(pred, gold)

    def values(self):
        out = {}
        for label, scorer in self.label__scorer.items():
            out[label] = scorer.f1()
        out['f1'] = sum(out.values()) / self.label_num
        return out


class ClassifyF1Scorer:
    # 计算多分类的F1
    def __init__(self, average='macro', labels=None):
        assert average in ['binary', 'micro', 'macro', 'weighted', 'samples'], f'average={average} 不合法的参数值'
        self.pred = []
        self.gold = []
        self.average = average
        self.labels = labels

    def update(self, pred, gold):
        if isinstance(pred, torch.Tensor):
            pred = pred.long().tolist()
        if isinstance(gold, torch.Tensor):
            gold = gold.long().tolist()
        self.pred.extend(pred)
        self.gold.extend(gold)

    def f1(self, reset=True):
        value = sklearn.metrics.f1_score(self.gold, self.pred, average=self.average, labels=self.labels)
        if reset:
            self.pred = []
            self.gold = []
        return value


class SetF1Scorer:
    # 通过预测和golden的集合计算F1
    def __init__(self):
        self.tp, self.fp, self.fn = 0, 0, 0

    def update(self, pred_set, gold_set):
        self.tp += len(pred_set & gold_set)
        self.fp += len(pred_set - gold_set)
        self.fn += len(gold_set - pred_set)

    def f1(self, reset=True):
        f1 = get_f1(self.tp, self.fp, self.fn)
        if reset:
            self.tp, self.fp, self.fn = 0, 0, 0
        return f1



import pymysql
from fish_tool.log_tool import logs


class MysqlTool:
    def __init__(self, user, password, database, host='127.0.0.1', port=3306, charset='utf8'):
        self.conn = pymysql.connect(host=host, port=port, user=user, password=password, database=database, charset=charset)

    def exec(self, sql, cursor=None):
        if not sql.endswith(';'):
            sql += ';'

        close = cursor is None
        cursor = cursor or self.conn.cursor()
        cursor.execute(sql)
        out = cursor.fetchall()
        if close:
            cursor.close()
        return out

    def count(self, table):
        sql = f'select count(*) from {table} limit 1;'
        resp = self.exec(sql)
        return resp[0][0]

    def keys(self, table):
        sql = f'show columns from {table};'
        resp = self.exec(sql)
        out = []
        for line in resp:
            one = {k: v for k, v in zip(['field', 'type', 'null', 'key', 'default', 'extra'], line)}
            # key: PRI-主键的组成部分    空-可以重复，没有索引     UNI-唯一值，不能为空    MUL-可以重复
            logs.rotate.info(one)
            out.append(one['field'])
        return out



import json
import sqlite3
from fish_tool.log_tool import logs, InfoError


class SqliteTool:
    def __init__(self, path):
        self.db = sqlite3.connect(path)

    def get_table_names(self):
        sql = "SELECT name FROM sqlite_master WHERE type='table' ORDER BY name;"
        cur = self.db.cursor()
        cur.execute(sql)
        lines = cur.fetchall()
        out = [t[0] for t in lines]
        return out

    def add_field(self, table, field, ftype, default):
        if ftype == 'int':
            ftype = 'integer'
        elif ftype == 'float':
            ftype = 'real'
        elif ftype == 'str':
            ftype = 'text'
        sql = f"alter table {table} add column '{field}' {ftype} default '{default}'"
        cur = self.db.cursor()
        out = cur.execute(sql)
        return out

    def create_table(self, table, field__type, uni_keys=None, primary_keys=None, auto_key=None):
        # uni_key 值必须唯一      primary_key 主键（便于搜索）   auto_key 自增key
        # field__type.type in ['str', 'int', 'float', 'blob']
        uni_keys = uni_keys or []
        primary_keys = primary_keys or []
        primary_keys = [t for t in primary_keys if t != auto_key]
        if primary_keys and auto_key:
            raise ValueError(f'如果有自增主键 则不能有其他主键 primary={primary_keys} auto={auto_key}')

        cur = self.db.cursor()
        field__type_txt = ''
        for k, v in field__type.items():
            if field__type_txt:
                field__type_txt += ', '
            if v == 'int':
                v = 'integer'
            elif v == 'float':
                v = 'real'
            elif v == 'str':
                v = 'text'
            if k in uni_keys:
                v = f'{v} UNIQUE'
            field__type_txt += f'"{k}" {v}'
        if primary_keys:
            keys = ', '.join(f'"{t}"' for t in primary_keys)
            field__type_txt += f', PRIMARY KEY({keys})'
        elif auto_key:
            field__type_txt += f', PRIMARY KEY("{auto_key}" AUTOINCREMENT)'

        sql = f'CREATE TABLE "{table}" ({field__type_txt});'
        cur.execute(sql)

    def num(self, table, k=None, v=None, condition=''):
        cur = self.db.cursor()
        sql = f'select count(*) from {table}'
        if condition:
            sql += f' where {condition}'
        elif k and v:
            if isinstance(v, str):
                v = f'"{v}"'
            condition = f'{k}={v}'
            sql += f' where {condition}'

        cur.execute(sql)
        lines = cur.fetchall()
        line = lines[0]
        out = line[0]
        return out

    def select(self, table, condition=None, num=None):
        cur = self.db.cursor()
        sql = f'select * from {table}'
        if condition:
            sql += f' where {condition}'
        if num is not None:
            sql += f' limit {num}'
        cur.execute(sql)
        fields = [t[0] for t in cur.description]
        out = []
        for row in cur.fetchall():
            obj = {}
            for k, v in zip(fields, row):
                if isinstance(v, bytes):
                    v = json.loads(v.decode())
                obj[k] = v
            out.append(obj)
        return out

    def commit(self):
        self.db.commit()

    def insert(self, table, field__value, commit=True):
        cur = self.db.cursor()
        fields = list(field__value)
        field_txt = ','.join(fields)
        ask_txt = ','.join('?' for _ in fields)
        values = []
        for k in fields:
            v = field__value[k]
            if not isinstance(v, (int, float, str)):
                v = json.dumps(v).encode()
            values.append(v)
        sql = f'insert into {table} ({field_txt}) values ({ask_txt})'
        cur.execute(sql, values)
        if commit:
            self.db.commit()

    def update(self, table, condition, field__value, commit=True):
        cur = self.db.cursor()
        fields = list(field__value)
        field_txt = ','.join(f'{t}=?' for t in fields)
        values = []
        for k in fields:
            v = field__value[k]
            if not isinstance(v, (int, float, str)):
                v = json.dumps(v).encode()
            values.append(v)
        sql = f'update  {table} set {field_txt} where {condition}'
        cur.execute(sql, values)
        if commit:
            self.db.commit()

    def delete(self, table, condition, commit=True):
        cur = self.db.cursor()
        sql = f'delete from {table} where {condition}'
        cur.execute(sql)
        if commit:
            self.db.commit()

    def exist(self, table, k, v):
        cur = self.db.cursor()
        sql = f"select exists(select 1 from {table} where {k}='{v}')"
        # log.info(f'sql={sql}')
        cur.execute(sql)
        resp = cur.fetchone()
        return resp[0]
        # if cur.fetchone():
        #     return True
        # else:
        #     return False


if __name__ == '__main__':
    db = SqliteTool('test.sqlite')



class GroupList:
    # 组列表 每个组由有关联的成员组成， 如果两个组的成员有关联则2个组合并成一个组（类似并查集）
    def __init__(self):
        self.item__team = {}  # 每个成员对应的team team={'items': set(), 'key': item}
        self.key_items = set()  # 每个team出一个关键成员， 便于迭代返回全部team

    def __bool__(self):
        if self.item__team:
            return True
        return False

    def __len__(self):
        return len(self.key_items)

    def __iter__(self):
        # 返回全部组的成员集合
        for key in self.key_items:
            yield self.item__team[key]['items']

    def __str__(self):
        txt = ', '.join(str(t) for t in self)
        return txt

    def __call__(self, item):
        # 返回item所属的组的成员集合 如果不存在则返回空集合
        if item in self.item__team:
            return self.item__team[item]['items']
        return set()

    def add(self, *items):
        # 添加相互关联的成员列表
        key_item = items[0]
        items = set(items)
        key__rel_idx_set = {}
        for item in items:
            if item in self.item__team:
                team = self.item__team[item]
                key = team['key']
                if key not in key__rel_idx_set:
                    key__rel_idx_set[key] = team
        if len(key__rel_idx_set) == 0:
            team = {'items': items, 'key': key_item}
            self.key_items.add(key_item)
            for item in items:
                self.item__team[item] = team
        elif len(key__rel_idx_set) == 1:
            teams = list(key__rel_idx_set.values())
            teams[0]['items'].update(items)
            for item in items:
                self.item__team[item] = teams[0]
        else:
            teams = list(key__rel_idx_set.values())
            key_team = teams[0]
            key_team['items'].update(items)
            for item in items:
                self.item__team[item] = key_team
            for team in teams[1:]:
                key_team['items'].update(team['items'])
                for item in team['items']:
                    self.item__team[item] = key_team
                self.key_items.remove(team['key'])



import random
import io
import types
import os
import copy
import math
import numpy as np
import sentencepiece
import torch
import torch.nn.parameter
from fish_tool.log_tool import logs, InfoError
from fish_tool.db import zip_tool

device = os.environ.get('CUDA_VISIBLE_DEVICES', '-1')


def list_tensor(*values, dtype=torch.long):
    return [tensor(v, dtype) for v in values]


def tensor(values, dtype=torch.long):
    # 把数组或者字典 转换成 tensor
    if not values:  # 空数组、空字典等等
        return values
    if isinstance(values, (list, np.ndarray)):
        return cuda(torch.tensor(values, dtype=dtype))
    elif isinstance(values, dict):
        return {k: tensor(v) for k, v in values.items()}
    else:
        return values


def count_to_weight(count, base=0.1):
    # 通过 统计数量 计算 损失加权
    max_num = max(count)
    if not max_num:
        return [1 for i in count]
    weight = [max_num / (t + base * max_num) for t in count]
    weight = cuda(torch.tensor(weight, dtype=torch.float))
    return weight


def init_linear(input_linear, seed=1337):
    """初始化全连接层权重
    """
    torch.manual_seed(seed)
    scope = np.sqrt(6.0 / (input_linear.weight.size(0) + input_linear.weight.size(1)))
    torch.nn.init.uniform_(input_linear.weight, -scope, scope)
    # nn.init.uniform(input_linear.bias, -scope, scope)
    if input_linear.bias is not None:
        input_linear.bias.data.zero_()


class WarmupOptim:
    # 带预热机制的优化器
    def __init__(self, params, lr=1e-3, warm_step=10000, warm_num=10, optim=None, **kv):
        optim = optim or torch.optim.SGD
        # warm_step学习率预热提升到100%需要的总步数     warm_num学习率变化的总次数
        if isinstance(params, types.GeneratorType):
            params = list(params)  # 生成器对象转换成数组， 避免第二次调用时报错
        self.optim_cls = optim
        self.params = params
        self.warm_step = warm_step
        self.kv = kv
        self.warm_frequency = warm_step // warm_num
        self.add_lr = lr / (warm_num + 1)
        self.lr = self.add_lr
        self.optim = self.get_optim()
        self.count = 1

    def get_optim(self):
        optimizer = self.optim_cls(self.params, lr=self.lr, **self.kv)
        return optimizer

    def zero_grad(self):
        self.optim.zero_grad()

    def step(self):
        self.optim.step()
        if self.count <= self.warm_step and self.count % self.warm_frequency == 0:
            self.lr += self.add_lr
            logs.tmp.info(f'update lr = {self.lr}')
            self.optim = self.get_optim()
        self.count += 1


class MovingAverage:
    # Exponentially Weighted Moving-Average 指数加权平均 指数加权滑动平均
    def __init__(self, beta=0.95):
        assert 0 < beta < 1
        self._value = None
        self.beta = beta

    @property
    def value(self):
        if self._value is None:
            return 0
        return self._value

    def update(self, value):
        if self._value is None:
            self._value = value
        else:
            new_value = self.beta * self._value + (1 - self.beta) * value
            if math.isnan(new_value):
                logs.tmp.error(f'value={value}  old={self._value}  new={new_value}')
            else:
                self._value = new_value


class SentencePieceVocab:
    def __init__(self, sp_path):
        if not os.path.isfile(sp_path):
            logs.brief.warning(f'sp_path文件不存在： {sp_path}')
        self.sp = sentencepiece.SentencePieceProcessor()
        self.sp.load(sp_path)

    def char_id(self, char):
        return self.sp.PieceToId(char)  # 如果piece不存在则输出是0

    def char(self, char_id):
        return self.sp.IdToPiece(char_id)

    def char_id_bak(self, char):
        pad_id = 6
        #  sp.EncodeAsPieces('萨')->['▁', '萨']   sp.encode_as_ids('阿')->[19, 893]
        this_encode = self.sp.encode_as_ids(char)
        if len(this_encode) >= 1:
            return this_encode[-1]
        else:
            return pad_id


def select_vec_3d(seq_vec, idxs):
    # seq_vec~(bsz, max_seq_len, dim)   idxs~(num)  idxs在dim=1上抽取向量
    # return~(bsz, num, dim)
    bsz, max_seq_len, dim = seq_vec.size()
    idxs = torch.tensor(idxs, dtype=torch.long, device=seq_vec.device)
    idxs = idxs.unsqueeze(0)
    idxs = idxs.unsqueeze(2)
    idxs = idxs.expand(bsz, -1, dim)
    return seq_vec.gather(1, idxs).contiguous()


def get_seq_combination_pair(seq_vec):
    # 输入向量序列 形成两两组合的pair向量序列
    # seq_vec~(bsz, max_seq_len, dim)
    # return: pair_vec~(bsz, join_num, dim)    join_num=max_seq_len*(max_seq_len-1)/2 两两成组
    max_seq_len = seq_vec.size(1)
    small_idxs = []  # (join_num)
    big_idxs = []  # (join_num)
    for j in range(1, max_seq_len):
        for i in range(j):
            small_idxs.append(i)
            big_idxs.append(j)
    left_vec = select_vec_3d(seq_vec, small_idxs)  # (bsz, join_num, dim)
    right_vec = select_vec_3d(seq_vec, big_idxs)  # (bsz, join_num, dim)
    pair_vec = torch.cat([left_vec - right_vec, left_vec * right_vec], dim=-1)  # (bsz, join_num, dim*2)
    return pair_vec


def __cuda(v):
    if torch.cuda.is_available() and hasattr(v, 'cuda'):
        return v.cuda()
    else:
        return v


def recusive__cuda(model):
    if isinstance(model, dict):
        return {k: recusive__cuda(v) for k, v in model.items()}
    elif isinstance(model, (list, tuple, set)):
        return [recusive__cuda(t) for t in model]
    else:
        return __cuda(model)


def cuda(*model):
    # 对数组参数的包装 调用代码省事一点
    if len(model) == 1:
        return recusive__cuda(model[0])
    else:
        return [recusive__cuda(t) for t in model]


def clone_module(module, num, share=False):
    if share:
        m = torch.nn.ModuleList([module for i in range(num)])
    else:
        m = torch.nn.ModuleList([copy.deepcopy(module) for i in range(num)])
    return m


def get_model_param_num(model):
    # 获取模型 参数量
    out = 0
    if isinstance(model, torch.nn.Module):
        out = sum(x.numel() for x in model.parameters())
    elif isinstance(model, torch.nn.parameter.Parameter):
        out = model.numel()
    return out


def get_model_layers_param_num(model):
    # 获取模型每层的参数量
    out = f'{model.__class__}  模型参数总量={get_model_param_num(model)}\n'
    for k, value in model.named_parameters():
        out += f'{k: <70} ---> {get_model_param_num(value)}\n'
    return out


def total_grad(model):
    # 查看一个模型的全部梯度之和
    out = 0
    for k, value in model.named_parameters():
        if isinstance(value, torch.nn.Module):
            out += total_grad(value)
        elif isinstance(value, torch.nn.parameter.Parameter):
            if value.grad is not None:
                out += value.grad.abs().sum()
    return out


def cut_grad(model, max_grad):
    # 对模型进行梯度剪裁 ，避免梯度更新过于剧烈
    torch.nn.utils.clip_grad_value_(model.parameters(), max_grad)


def count_to_lr_weight(counts, base=0.05, device=None):
    big = max(counts)
    add = big * base  # 平滑
    counts_pad = [t + add for t in counts]
    big += add
    weight = [big / t for t in counts_pad]
    weight = torch.tensor(weight, dtype=torch.float, device=device)
    return weight


def save_model(model, path):
    path = os.path.abspath(path)
    save_dir = os.path.dirname(path)
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    torch.save(model.state_dict(), path)
    return model


def load_cpu_model(model, path):
    if isinstance(path, bytes):
        # 如果是二进制 则已经读取了 要包装一下
        path = io.BytesIO(path)
    model_state = torch.load(path, map_location=torch.device('cpu'))
    model.load_state_dict(model_state)
    return model


def model_save_zip(model, path, name='model.pth'):
    # 把模型存储到zip文件中 注意存储过程很慢
    save_dir = os.path.dirname(os.path.abspath(path))
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    zip = zip_tool.ZipTool(path)
    if zip.is_exist(name):
        raise InfoError(f'模型存储错误 path={path} name={name} 已经存在')
    f = io.BytesIO()
    torch.save(model.state_dict(), f)
    pth = f.getvalue()
    zip.add(name, pth)


def model_load_zip(model, path, name='model.pth'):
    zip = zip_tool.ZipTool(path, 'r')
    pth = zip.read(name)
    model = load_cpu_model(model, pth)
    return model


def __init_model_weights(module, initializer_range=0.02, seed=1337):
    torch.manual_seed(seed)
    if isinstance(module, torch.nn.Linear):
        scope = math.sqrt(6.0 / (module.weight.size(0) + module.weight.size(1)))
        torch.nn.init.uniform_(module.weight, -scope, scope)
        if module.bias is not None:
            module.bias.data.zero_()
    elif isinstance(module, torch.nn.Embedding):
        module.weight.data.normal_(mean=0.0, std=initializer_range)
        if isinstance(module, torch.nn.Linear) and module.bias is not None:
            module.bias.data.zero_()
    elif isinstance(module, torch.nn.LayerNorm):
        module.bias.data.zero_()
        module.weight.data.fill_(1.0)


def init_model_weights(model):
    model.apply(__init_model_weights)


def ce_loss(pred, target, weight=None, mask=None):
    # pred~(bsz, seq_len, cls_num)   target~(bsz, seq_len)
    # 安全的交叉熵损失, 避免出现NaN的情况
    target = torch.nn.functional.one_hot(target, num_classes=pred.size(-1)).float()  # ~ (bsz, seq_len, cls_num)
    pred = torch.softmax(pred, dim=-1)  # ~ (bsz, seq_len, cls_num)
    pred = pred.clamp(min=1e-20, max=1)
    shape_loss = target * torch.log(pred)  # ~ (bsz, seq_len, cls_num)
    if mask is not None:
        mask = mask.float()
        shape_loss = shape_loss * mask
    shape_loss = shape_loss.sum(dim=-1)  # ~ (bsz, seq_len)
    if weight is not None:
        shape_loss = shape_loss * weight
    loss = -torch.mean(shape_loss)
    return loss


def ce_loss_with_probability(pred, target, weight=None, mask=None):
    # pred~(bsz, seq_len)二分类的概率值   target~(bsz, seq_len)取值只能是0或1
    target = target.float()
    pred_1 = (1 - pred).clamp(min=1e-20, max=1)
    pred = pred.clamp(min=1e-20, max=1)
    if weight is None:
        weight0, weight1 = 1, 1
    else:
        weight0, weight1 = weight
    shape_loss = -(1 - target) * torch.log(pred_1) * weight0 - target * torch.log(pred) * weight1
    loss = torch.mean(shape_loss)
    return loss


def pad_1d(line, max_num, pad=0):
    # line是1维数组， 补全到max_num长度
    out = line + [pad] * (max_num - len(line))
    return out


def pad_batch_seq(batch_seq, pad=0):
    # batch_seq 是2维数组， (batch_size, seq_len)
    max_len = max(len(seq) for seq in batch_seq)
    mask = []
    input_ids = []
    for seq in batch_seq:
        input_ids.append(pad_1d(seq, max_len, pad=pad))
        mask.append([1] * len(seq) + [0] * (max_len - len(seq)))
    return input_ids, mask


def pad_2d(matrix, max_row, max_column, pad=0):
    # matrix是2维数组， 补全到max-row行数， max-column列数
    row = len(matrix)
    if row:
        column = len(matrix[0])
    else:
        # 如果原矩阵是空白矩阵，直接返回pad组成的对应尺寸的矩阵
        return [[pad] * max_column for _ in range(max_row)]
    out = []
    for line in matrix:
        out.append(line + [pad] * (max_column - column))
    out += [[pad] * max_column for _ in range(max_row - row)]
    return out


def frozen(net):
    # 冻结网络的参数
    for param in net.parameters():
        param.requires_grad = False


def unfrozen(net):
    # 解冻网络的参数
    for param in net.parameters():
        param.requires_grad = True


def graph_to_cliques(nodes, edges):
    """
    输入一个无向图（顶点和边），解码出这个图全部的团
    :param nodes: 顶点数组   range(1, 6)
    :param edges: 边数组, 每个边由二元组表示  {(1, 2), (1, 4), (1, 5), (2, 3), (2, 5), (3, 5), (4, 5)}
    :return: 顶点的二维数组，表示每个团    [[1, 2, 5], [1, 4, 5], [2, 3, 5]]
    """
    nodes = sorted(nodes)
    node_relation_nodes = {}
    for a, b in edges:
        node_relation_nodes.setdefault(a, set()).add(b)
        node_relation_nodes.setdefault(b, set()).add(a)

    cliques = []
    looked_nodes = set()
    for node in nodes:
        relation_nodes = node_relation_nodes.get(node, set()) & looked_nodes
        __clique_add_node(cliques, node, relation_nodes)
        looked_nodes.add(node)
    cliques = [sorted(t) for t in cliques]
    cliques.sort()
    return cliques


def __clique_add_node(cliques, node, relation_nodes):
    # 操作 cliques 把节点加入到 已有的团， 或者新建一个团
    if not cliques:  # 如果已创建的团为空
        cliques.append({node})
        # logs.tmp.debug(f'\t创建一个团  已有cliques={cliques}')
        return
    if not relation_nodes:  # 当前节点没有任何边
        cliques.append({node})
        # logs.tmp.debug(f'\t创建一个团  已有cliques={cliques}')
        return
    share_vertexes = set()  # 当前节点加入的团的全部节点
    for i, clique in enumerate(cliques):
        if relation_nodes == clique:
            share_vertexes.update(clique)
            clique.add(node)
            # logs.tmp.debug(f'\t完整加入  已有cliques={cliques}')
        elif clique.issubset(relation_nodes):
            share_vertexes.update(clique)
            clique.add(node)
            # logs.tmp.debug(f'\t共享加入  已有cliques={clique}')
    left_vartexes = relation_nodes - share_vertexes  # 剩下的，没有加入到已有团里的边的节点
    if left_vartexes:
        left_vartexes.add(node)
        cliques.append(left_vartexes)
        # logs.tmp.debug(f'\t共享并创建一个团  已有cliques={cliques}')
    return


def pair_to_objs(ent_pair, ent_num):
    # 根据实体对是否属于同一个实体（边），解码出实体成组（最大团）
    if ent_num == 0:
        return []
    elif ent_num == 1:
        return [[0]]
    nodes = range(ent_num)
    edges = set()
    i = 0
    for big in range(1, ent_num):
        for small in range(big):
            if ent_pair[i] > 0:
                edges.add((small, big))
            i += 1
    cliques = graph_to_cliques(nodes, edges)
    return cliques


def decode_ner_line(tags):
    """
    :param tags: 数字 O=out  1=Begin  2=inner  3=end  4=single
    return: [(start, end)]  index左闭右开
    """
    out = []
    start = None
    for i, tag in enumerate(tags):
        if tag == 0:
            if start is not None:
                out.append((start, i))
                start = None
        elif tag == 1:
            if start is not None:
                out.append((start, i))
            start = i
        elif tag == 2:
            if start is None:
                start = i
        elif tag == 3:
            if start is None:
                start = i
            out.append((start, i + 1))
            start = None
        elif tag == 4:
            if start is not None:
                out.append((start, i))
            start = None
            out.append((i, i + 1))
    if start is not None:
        out.append((start, len(tags)))
    return out


def seed_everything(seed):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


def randn(*size, variance=1.0, clamp=20.0, requires_grad=False):
    # 创建正态分布的张量
    vec = torch.randn(*size, requires_grad=requires_grad)
    vec = vec * variance
    assert clamp > 0
    vec.clamp_(min=-clamp, max=clamp)
    return vec


if __name__ == '__main__':
    # 测试pytorch计算精度不一致问题
    logit = torch.FloatTensor([0.621, 0.4067])
    soft = logit.softmax(-1)
    prob0 = 1 - soft[1]
    print(prob0.tolist(), soft.tolist())
    # (0.5533709526062012, [0.5533708930015564, 0.4466290771961212])

    ce_fn = torch.nn.CrossEntropyLoss()
    gold = torch.LongTensor([[0]])
    pred = torch.FloatTensor([[[0.621, 0.4067]]])
    pred = torch.FloatTensor([[[0.621, 0.6067]]])
    print(f'pred={pred.tolist()}')
    logit = pred.softmax(dim=-1)
    print(f'logit={logit}')
    prob = logit[:, :, 1]
    print(f'prob={prob}')
    print(f'ce_fn={ce_fn(pred.view(-1, 2), gold.view(-1))}')
    print(f'my_ce={ce_loss(pred, gold)}')
    print(f'ce_pr={ce_loss_with_probability(prob, gold)}')

    x = torch.FloatTensor([[0.6860]])
    xx = torch.mean(x)
    logs.tmp.info(f'x={x}  mean={xx}')

    x = torch.FloatTensor([[-0.6860]])
    xx = -torch.mean(x)
    logs.tmp.info(f'x={x}  mean={xx}')



import os
import zipfile
from fish_tool.log_tool import logs
from fish_tool import sys_tool

default_compress_type = zipfile.ZIP_LZMA  # zipfile.ZIP_BZIP2   zipfile.ZIP_DEFLATED


class ZipTool:
    def __init__(self, path, mode=None, compress_type=None):
        if mode is None:
            if os.path.exists(path):
                mode = 'a'  # r只读  w创建  x独占创建  a后续添加
                logs.brief.info(f'自动指定 mode=a 文件已经存在 {path}')
            else:
                mode = 'w'
                logs.brief.info(f'自动指定 mode=w 文件不存在 {path}')
        os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)
        self.compress_type = default_compress_type if compress_type is None else compress_type
        self.f = zipfile.ZipFile(path, mode, self.compress_type)

    def add(self, name, content):
        # content 可以是二进制文件 也可以是文本内容
        with sys_tool.DelayKeyboardInterrupt():
            self.f.writestr(name, data=content, compress_type=self.compress_type, compresslevel=None)

    def namelist(self):
        return self.f.namelist()

    def read(self, name):
        info = self.f.NameToInfo[name]  # 获取name对应得最新的 zipinfo（同一个name可以多次写入数据）
        return self.f.read(info)

    def open(self, name, mode="r", pwd=None, force_zip64=False):
        return self.f.open(name, mode=mode, pwd=pwd, force_zip64=force_zip64)

    def is_exist(self, name):
        return name in self.f.NameToInfo



import os
from fish_tool.log_tool import logs, InfoError


class BaseConfig:

    def load(self, path: str):
        import yaml
        data = yaml.load(open(path, encoding='utf-8'), Loader=yaml.loader.Loader)
        self.__dict__.update(data)
        logs.tmp.info(f'load config: {self}  path={path}')
        return self

    def __getattr__(self, item):  # 未定义的属性返回None 避免引起异常
        return None

    def all_kv(self):
        kv = {k: v for k, v in self.__class__.__dict__.items() if not k.startswith('_')}  # 不保存以下划线开头的变量名
        kv.update(self.__dict__)
        return kv

    def save(self, path):
        import yaml
        path_dir = os.path.dirname(path)
        os.makedirs(path_dir, exist_ok=True)
        logs.tmp.info(f'save config: {path}')
        with open(path, 'w', encoding='utf-8') as f:
            f.write(yaml.dump(self.all_kv(), allow_unicode=True))
        return self

    def __str__(self):
        return f'{self.__class__.__name__}({self.all_kv()})'

    def check(self, *keys):
        # 检查config是否包含 这些字段
        kv = self.all_kv()
        error_keys = [k for k in keys if k not in kv]
        if error_keys:
            raise InfoError(f'config 缺少字段 {error_keys}')



from setuptools import setup, find_packages

setup(
    name="fish_tool",
    version="1.1",
    author="aifish",
    url="https://gitee.com/laowangzi/fish_tool",
    author_email="ofyu@163.com",
    description="some tool code",
    packages=find_packages(),
    install_requires=[
        'pyyaml>=6.0',
        'tqdm>=4.36.1',
        'requests>=2.22.0',
        'httpx>=0.18.1',
        "fastapi>=0.90.0",
        "uvicorn>=0.20.0",
    ],
    extras_require={
        'AI': [
            "torch>=1.0"
        ],
        'db': [
            "pymysql>=1.0.2"
        ],
    },
    python_requires='>=3.6',
)

"""
build代码
python setup.py dist

上传到PyPi
(pip install twine)
twine upload dist/*
"""
"""
更新内容
1.01 2023.03.22 db/sqlite_tool.py 增加字段功能
"""
